r3d=require "reactphysics3d"

local world=r3d.World.new(0,-4,0)
local sw,sh=application:getContentWidth(),application:getContentHeight()
local view=D3.View.new(sw,sh,45,0.1,1000)

-- Build a generic CUBE Mesh with normals
local GENCUBE=D3.Cube.new(1.2, 1.2, 1.2) 
GENCUBE:mapTexture(Texture.new("box.png",true))
GENCUBE:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW)
GENCUBE.shape=r3d.BoxShape.new(1.2, 1.2, 1.2)

-- Build a 100x100 floor plane (with normals)
local gplane=D3.Mesh.new()
gplane:setVertexArray{-100,1,-100, 100,1,-100, 100,1,100, -100,1,100}
local tw,th=3200,3200 
gplane:setTextureCoordinateArray{0,0,tw,0,tw,th,0,th}
gplane:setGenericArray(3,Shader.DFLOAT,3,4,{ 0,1,0,0,1,0,0,1,0,0,1,0, })
gplane:setIndexArray{1,2,3,1,3,4}
gplane:setTexture(Texture.new("grass.png",true,{wrap=TextureBase.REPEAT}))
gplane:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW|D3.Mesh.MODE_TEXTURE)
gplane.body=world:createBody()
gplane.body:setType(r3d.Body.STATIC_BODY)
gplane.body:createFixture(r3d.BoxShape.new(100,1,100),nil,1000)

--Function to build a cube instance
world.bodies={}
function build_cube()
	-- We build a viewport onto the generic cube we set up earlier
	local v=Viewport.new()
	v:setContent(GENCUBE)
	-- And randomly place/rotate it
	v:setPosition(math.random()*9-5,math.random()*10+10,math.random()*9-5)
	v:setRotation(math.random()*20-10)
	v:setRotationX(math.random()*20-10)
	v:setRotationY(math.random()*20-10)
	-- Create body accordingly
	local body=world:createBody(v:getMatrix())
	body:createFixture(GENCUBE.shape,nil,1)
	-- Add the body to our moving bodies list
	world.bodies[v]=body
	return v
end

-- two boxes
local box01=Viewport.new()
box01:setContent(GENCUBE)
box01:setPosition(0, 2, 0)
box01.body=world:createBody(box01:getMatrix())
box01.body:createFixture(GENCUBE.shape,nil,1)
box01.body.isdirty = false

local box02=Viewport.new()
box02:setContent(GENCUBE)
box02:setPosition(0.5, 7, 0)
box02.body=world:createBody(box02:getMatrix())
box02.body:createFixture(GENCUBE.shape,nil,1)
box02.body.isdirty = false

--Setup our scene
local scene=view:getScene()
scene:addChild(gplane)
scene:addChild(box01)
scene:addChild(box02)
local debugDraw = r3d.DebugDraw.new(world)
scene:addChild(debugDraw)

--Configure light
Lighting.setLight(15,30,0,0.3)
Lighting.setLightTarget(0,0,0,40,120)

--Setup camera
view:lookAt(0,10,-20,0,5,0)

stage:addChild(view)

-- various collision detection
function box01collision()
	print("box01 collision")
	box01.body.isdirty = true
end

function box02collision()
--	print("box02 collision")
	box02.body.isdirty = true
end

-- Main loop: 
local gen=0
stage:addEventListener(Event.ENTER_FRAME,function(e)
	-- destroy body before world step?
	if box02.body.isdirty then
		box02.body.isdirty = false
		print("box02 is dirty is true, try to destroy it!")
--		world:destroyBody(box02.body) -- CRASH!
	end
	-- generate cubes
	gen=gen+1
	if gen==200 then
		scene:addChild(build_cube())	
		gen=0
	end
	-- world step
	world:step(e.deltaTime)
	if box01.body ~= nil then
		box01:setMatrix(box01.body:getTransform())
	end
	if box02.body ~= nil then
		box02:setMatrix(box02.body:getTransform())
	end
	for s,b in pairs(world.bodies) do
		s:setMatrix(b:getTransform())
	end
	-- collisions
	if box01.body ~= nil then
		world:testCollision(box01.body, box01collision) -- OK
	end
	if box02.body ~= nil then
		world:testCollision(box02.body, box02collision) -- OK
	end
	-- lighting
	Lighting.computeShadows(scene)
end)
